import os
import pytest
from cymseed3 import MiniSeed

# Path to the directory containing example MiniSEED files
DATA_DIR = os.path.join(os.path.dirname(__file__), "data")


@pytest.fixture
def miniseed():
    """Fixture for initializing and freeing MiniSeed object."""
    ms = MiniSeed()
    yield ms
    del ms


def test_import_record(miniseed):
    """Test importing a record into MiniSeed."""
    record_dict = {
        'network': 'OX',
        'station': 'ABCD',
        'location': '00',
        'channel': 'HHZ',
        'starttime': '2010-01-10T08:23:45.019538Z',
        'rate': 200,
        'nsamp': 100,
        'data': [i for i in range(100)]
    }
    miniseed.import_record(record_dict)
    assert miniseed is not None, "Failed to import record into MiniSeed"


def test_write_and_read(miniseed, tmp_path):
    """Test writing and reading MiniSEED records to and from a file."""
    record_dict = {
        'network': 'OX',
        'station': 'ABCD',
        'location': '00',
        'channel': 'HHZ',
        'starttime': '2010-01-10T08:23:45.019538Z',
        'rate': 200,
        'nsamp': 100,
        'data': [i for i in range(100)]
    }
    miniseed.import_record(record_dict)
    
    # Write data into binary buffer
    byte_buffer = miniseed.write(msformat=2, reclen=512, encoding=11)
    assert byte_buffer, "Failed to write MiniSEED data to buffer"

    # Write bytes to a temporary file
    output_file = tmp_path / "my_file.mseed"
    with open(output_file, "wb") as fid:
        fid.write(byte_buffer)
    assert output_file.exists(), "Failed to write buffer to file"

    # Read the file
    with open(output_file, 'rb') as fid:
        byte_buffer = fid.read()
    assert byte_buffer, "Failed to read MiniSEED data from file"

    # Read the buffer back into the MiniSeed object
    miniseed.read(byte_buffer, verbose=1)
    rec_list = miniseed.export_records()
    assert len(rec_list) > 0, "Failed to export records from MiniSeed"


def test_export_records(miniseed):
    """Test exporting records from MiniSeed to a list of dictionaries."""
    record_dict = {
        'network': 'OX',
        'station': 'ABCD',
        'location': '00',
        'channel': 'HHZ',
        'starttime': '2010-01-10T08:23:45.019538Z',
        'rate': 200,
        'nsamp': 100,
        'data': [i for i in range(100)]
    }
    miniseed.import_record(record_dict)
    rec_list = miniseed.export_records()
    assert isinstance(rec_list, list), "Exported records are not a list"
    assert len(rec_list) > 0, "Exported records list is empty"
    assert rec_list[0]['network'] == 'OX', "Record data mismatch"
    assert rec_list[0]['station'] == 'ABCD', "Record data mismatch"
    assert rec_list[0]['location'] == '00', "Record data mismatch"
    assert rec_list[0]['channel'] == 'HHZ', "Record data mismatch"
    assert rec_list[0]['starttime'] == '2010-01-10T08:23:45.019538Z', \
        "Record data mismatch"
    assert rec_list[0]['rate'] == 200, "Record data mismatch"
    assert rec_list[0]['nsamp'] == 100, "Record data mismatch"
    assert rec_list[0]['data'] == [i for i in range(100)], \
        "Record data mismatch"



# ****************************************************************************
#
# Copyright (C) 2019-2024, ShakeLab Developers.
# This file is part of ShakeLab.
#
# ShakeLab is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# ShakeLab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# with this download. If not, see <http://www.gnu.org/licenses/>
#
# ****************************************************************************
"""
MiniSEED Processing Module

This module provides the `MiniSeed` class and associated utility functions 
for processing MiniSEED data using the `libmseed` library. It allows for 
the reading, writing, and manipulation of MiniSEED data, as well as the 
conversion of time formats and encoding/decoding of records.

Note on Encoding:
    0 -> Text encoding (UTF-8)
    1 -> 16-bit integer
    3 -> 32-bit integer
    4 -> 32-bit float (IEEE)
    5 -> 64-bit float (IEEE)
    10 -> Steim-1 compressed integers
    11 -> Steim-2 compressed integers
    12 -> [Legacy] GEOSCOPE 24-bit integer
    13 -> [Legacy] GEOSCOPE 16-bit gain ranged, 3-bit exponent
    14 -> [Legacy] GEOSCOPE 16-bit gain ranged, 4-bit exponent
    16 -> [Legacy] CDSN 16-bit gain ranged
    30 -> [Legacy] SRO 16-bit gain ranged
    32 -> [Legacy] DWWSSN 16-bit gain ranged
"""

# cymseed3.pyx
# cython: language_level=3

cimport cymseed3
from cython.operator import dereference

from libc.string cimport memset, memcpy
from libc.stdlib cimport malloc, realloc, free

# -------------------------------------------------------------------------- #
# TIME CONVERSION FUNCTIONS
# -------------------------------------------------------------------------- #

# Define constants for time format and subsecond precision
TIMEFORMAT_ISO = 1 
SUBSECOND_MICRO = 3

def convert_components_to_nstime(int year, int yday, int hour,
                                 int min, int sec, uint32_t nsec):
    """
    Convert time components to nstime_t (nanoseconds since epoch).
    
    Parameters:
        year (int): The year component.
        yday (int): The day of the year component.
        hour (int): The hour component.
        min (int): The minute component.
        sec (int): The second component.
        nsec (uint32_t): The nanosecond component.
    
    Returns:
        nstime_t: The time in nanoseconds since epoch.
    """
    return ms_time2nstime(year, yday, hour, min, sec, nsec)

def convert_timestr_to_nstime(str timestr):
    """
    Convert a formatted time string to nstime_t.
    
    Parameters:
        timestr (str): The ISO formatted time string.
    
    Returns:
        nstime_t: The time in nanoseconds since epoch.
    """
    return ms_timestr2nstime(timestr.encode('utf-8'))

def convert_nstime_to_components(int64_t nstime):
    """
    Convert nstime_t to year, day of year, hour,
    minute, second, and nanosecond components.
    
    Parameters:
        nstime_t (int64_t): The time in nanoseconds since epoch.
    
    Returns:
        tuple: A tuple containing year, day of year, hour,
               minute, second, nanosecond.
    """
    cdef uint16_t year, yday
    cdef uint8_t hour, min, sec
    cdef uint32_t nsec
    
    ms_nstime2time(nstime, &year, &yday, &hour, &min, &sec, &nsec)
    
    return year, yday, hour, min, sec, nsec

def convert_nstime_to_timestr(int64_t nstime):
    """
    Convert nstime_t to a formatted ISO time string.
    
    Parameters:
        nstime_t (int64_t): The time in nanoseconds since epoch.
    
    Returns:
        str: The ISO formatted time string.
    """
    cdef char timestr[30]
    
    cdef int timeformat_iso = TIMEFORMAT_ISO
    cdef int subsecond_micro = SUBSECOND_MICRO
    
    ms_nstime2timestr(nstime, timestr, timeformat_iso, subsecond_micro)
    
    return timestr.decode('utf-8')

# -------------------------------------------------------------------------- #

cdef void record_handler(char *record,
                         int reclen,
                         void *handlerdata) noexcept:
    """
    Handle each MiniSEED record during the packing process.

    This function is called by `mstl3_pack` for each MiniSEED record.
    It appends the record data to a binary buffer (bytearray)
    passed as `handlerdata`.

    Parameters:
        record (char*): Pointer to the MiniSEED record data.
        reclen (int): Length of the MiniSEED record.
        handlerdata (void*): Pointer to the handler data,
                             which is expected to be a bytearray.

    Notes:
        - The `noexcept` specifier indicates that this function
          does not raise Python exceptions.
        - The handler assumes that `handlerdata` is a bytearray
          that can be safely extended.
    """
    # For debug
    # cdef MS3Record* msr = NULL
    # if not msr3_parse (record, reclen, &msr, 0, 0):
    #     msr3_print (msr, 0)
    # msr3_free(&msr)

    # Cast the handlerdata to a bytearray to append the record data
    buffer = <bytearray> handlerdata

    # Ensure the record is valid and has a positive length before appending
    if record != NULL and reclen > 0:
        buffer.extend(<bytes>record[:reclen])

# -------------------------------------------------------------------------- #
# MINISEED MAIN CLASS
# -------------------------------------------------------------------------- #

cdef class MiniSeed:
    """
    MiniSeed class for handling MiniSEED data.
    
    This class provides functionality to initialize, read, write, export, 
    and import MiniSEED records using the libmseed library.
    """
    cdef MS3TraceList* mstl

    def __cinit__(self):
        """
        Initialize the MiniSeed class by creating an MS3TraceList.
        """
        self.mstl = mstl3_init(NULL)

    def __dealloc__(self):
        """
        Free the MS3TraceList when the MiniSeed object is deallocated.
        """
        mstl3_free(&self.mstl, 0)

    @property
    def numtraceids(self):
        return self.mstl.numtraceids

    def print(self):
        pass

    def read(self, bytes byte_stream,
                   uint32_t flags=0,
                   int8_t verbose=0):
        """
        Read and parse MiniSEED data from a byte stream.

        Parameters:
            byte_stream (bytes): The MiniSEED data to be parsed.
            flags (uint32_t): Flags to control parsing behavior.
            verbose (int8_t): Verbosity level for logging.
        """
        cdef MS3Record* msr
        cdef MS3RecordPtr* ptr

        cdef uint64_t bufferlength = len(byte_stream)
        cdef uint64_t offset = 0

        cdef int8_t splitversion = 0
        cdef int parsevalue

        # Set flags for validation, unpacking, and record listing
        flags |= MSF_VALIDATECRC
        flags |= MSF_UNPACKDATA
        flags |= MSF_SKIPNOTDATA
        flags |= MSF_RECORDLIST

        # Parse through the byte stream, creating and adding MS3Records
        while int(bufferlength - offset) > int(MINRECLEN):

            msr = msr3_init(NULL)
            if not msr:
                raise MemoryError("Failed to allocate memory for MS3Record")

            parsevalue = msr3_parse(byte_stream[offset:],
                                    bufferlength,
                                    &msr,
                                    flags,
                                    verbose)

            # Verbose logging of parsed record details
            if verbose > 0:
                iso_starttime = convert_nstime_to_timestr(msr.starttime)
                strout = f"SID: {msr.sid.decode('utf-8')}, "
                strout += f"Start Time: {iso_starttime}, "
                strout += f"Sample Count: {msr.samplecnt}"
                print(strout)

            if parsevalue < 0:
                if msr:
                    msr3_free(&msr)
                raise RuntimeError(f"Error parsing MiniSEED data")

            if parsevalue > 0:
                break

            # Add the parsed MS3Record to the trace list
            mstl3_addmsr_recordptr(self.mstl,
                                   msr,
                                   &ptr,
                                   splitversion,
                                   1,
                                   flags,
                                   NULL)

            offset += msr.reclen

    def write(self, int msformat=2,
                    int reclen=512,
                    int8_t encoding=11,
                    uint32_t flags=0,
                    int8_t verbose=0):
        """
        Write the trace list to a MiniSEED byte stream.

        Parameters:
            msformat (int): MiniSEED format version (2 or 3).
            reclen (int): Record length for the MiniSEED data.
            encoding (int8_t): Encoding format for the data.
            flags (uint32_t): Flags to control writing behavior.
            verbose (int8_t): Verbosity level for logging.

        Returns:
            bytes: The packed MiniSEED data.
        """
        cdef int64_t psamples
        cdef int precords

        byte_stream = bytearray()
    
        # Set flags for validation, unpacking, maintaining trace list,
        # and flushing data
        flags |= MSF_VALIDATECRC
        flags |= MSF_UNPACKDATA
        flags |= MSF_MAINTAINMSTL
        flags |= MSF_FLUSHDATA

        # Set miniSEED version 2 if requested
        if msformat == 2:
            flags |= MSF_PACKVER2

        # Pack the trace list into MiniSEED format
        precords = mstl3_pack(self.mstl,
                              record_handler,
                              <void *> byte_stream,
                              reclen,
                              encoding,
                              &psamples,
                              flags,
                              verbose,
                              NULL)

        return bytes(byte_stream)

    def export_records(self):
        """
        Convert the parsed MiniSEED data into a list of dictionaries.

        Each dictionary contains the NSLC codes, start time, sample rate,
        number of samples, and the data array for a segment.

        Returns:
            list: A list of dictionaries representing the MiniSEED records.
        """
        cdef list record_list = []
        traceid = self.mstl.traces

        # Prepare buffers for NSLC components
        cdef char network[10]
        cdef char station[10]
        cdef char location[10]
        cdef char channel[10]

        cdef int *int_ptr
        cdef float *float_ptr
        cdef double *double_ptr
        cdef char *char_ptr

        while traceid.next[0] != NULL:
            traceid = dereference(traceid.next[0])

            seg = traceid.first
            while seg:

                rng_cnt = range(seg.samplecnt)

                # Parse the SID into FSDN codes
                ms_sid2nslc(traceid.sid, network, station, location, channel)

                # Process the data samples based on the sample type
                if seg.sampletype == b'i':
                    int_ptr = <int *>seg.datasamples
                    data = [int_ptr[i] for i in rng_cnt]

                elif seg.sampletype == b'f':
                    float_ptr = <float *>seg.datasamples
                    data = [float_ptr[i] for i in rng_cnt]

                elif seg.sampletype == b'd':
                    double_ptr = <double *>seg.datasamples
                    data = [double_ptr[i] for i in rng_cnt]

                elif seg.sampletype == b't':
                    char_ptr = <char *>seg.datasamples
                    data = [char_ptr[i].decode('utf-8') for i in rng_cnt]

                else:
                    raise ValueError(f"Unsupported sampletype")

                # Create a record dictionary
                record_dict = {
                    'network': network.decode('utf-8'),
                    'station': station.decode('utf-8'),
                    'location': location.decode('utf-8'),
                    'channel': channel.decode('utf-8'),
                    'starttime': convert_nstime_to_timestr(seg.starttime),
                    'rate': seg.samprate,
                    'nsamp': seg.numsamples,
                    'data' : data
                }
                record_list.append(record_dict)
                seg = seg.next

        return record_list

    def import_record(self, record_dict,
                      int16_t encoding=11,
                      int32_t reclen=512,
                      int8_t verbose=0):
        """
        Import a record from a dictionary and add it to the trace list.

        Parameters:
            record_dict (dict): Dictionary containing record details
                (NSLC codes, start time, sample rate, number of samples).
            encoding (int16_t): Encoding format for the data.
            reclen (int32_t): Record length for the MiniSEED data.
            verbose (int8_t): Verbosity level for logging.
        """
        cdef MS3Record *msr
        cdef MS3RecordPtr* ptr
        cdef char sid[64]
        
        cdef int8_t splitversion = 0
        cdef uint32_t flags=0

        # Set flags for validation, unpacking, and record listing
        flags |= MSF_VALIDATECRC
        flags |= MSF_UNPACKDATA
        flags |= MSF_SKIPNOTDATA
        flags |= MSF_RECORDLIST

        # Initialize a new MiniSEED record
        msr = msr3_init(NULL)
        if not msr:
            raise MemoryError("Failed to allocate memory for MS3Record")

        # Define FDSN codes from the record dictionary
        network = record_dict['network'].encode('utf-8')
        station = record_dict['station'].encode('utf-8')
        location = record_dict['location'].encode('utf-8')
        channel = record_dict['channel'].encode('utf-8')

        cdef char* net = <char*> network  # Network code
        cdef char* sta = <char*> station  # Station code
        cdef char* loc = <char*> location # Location code
        cdef char* chn = <char*> channel  # Channel code

        # Generate SID using the FDSN codes
        ms_nslc2sid(sid, sizeof(sid), 0, net, sta, loc, chn)
        
        # Assign the generated SID to msr.sid
        msr.sid = sid

        # Set the record properties from the dictionary
        msr.starttime = convert_timestr_to_nstime(record_dict['starttime'])
        msr.samprate = <double> record_dict['rate']
        msr.samplecnt = <int64_t> record_dict['nsamp']
        msr.numsamples = msr.samplecnt

        msr.encoding = encoding
        msr.reclen = reclen

        # Allocate memory for data samples
        cdef int32_t* data = <int32_t*> malloc(msr.samplecnt * sizeof(int32_t))
        if not data:
            msr3_free(&msr)
            raise MemoryError("Failed to allocate memory for data samples")

        # Generate example data
        for i in range(msr.samplecnt):
            data[i] = record_dict['data'][i]

        # Set the sample data and its size in the MS3Record structure
        msr.datasamples = <void*> data

        # Determine the sample type and data size based on encoding
        if encoding in [0]:
            msr.sampletype = b't'
            msr.datasize = msr.samplecnt * sizeof(char)

        elif encoding in [1]:
            msr.sampletype = b'i'
            msr.datasize = msr.samplecnt * sizeof(int16_t)

        elif encoding in [4]:
            msr.sampletype = b'f'
            msr.datasize = msr.samplecnt * sizeof(float)

        elif encoding in [5]:
            msr.sampletype = b'd'
            msr.datasize = msr.samplecnt * sizeof(double)

        elif encoding in [3, 10, 11]:
            msr.sampletype = b'i'
            msr.datasize = msr.samplecnt * sizeof(int32_t)

        else:
            raise ValueError(f"Unsupported encoding")

        # NOTE: is it needed to pack data?

        # Add the record to the trace list
        mstl3_addmsr_recordptr(self.mstl,
                               msr,
                               &ptr,
                               splitversion,
                               1,
                               flags,
                               NULL)

        # Free the MS3Record
        msr3_free(&msr)

# -------------------------------------------------------------------------- #




from setuptools import setup, Extension
from Cython.Build import cythonize
import os

# Directory of the C library
libmseed_dir = os.path.join('earthscope', 'libmseed-3.1.3')

# Read the README file for the long description
long_description = ""
if os.path.exists('README.md'):
    with open('README.md', 'r') as f:
        long_description = f.read()

# List of C source files for libmseed
libmseed_sources = [
    os.path.join(libmseed_dir, 'crc32c.c'),
    os.path.join(libmseed_dir, 'extraheaders.c'),
    os.path.join(libmseed_dir, 'fileutils.c'),
    os.path.join(libmseed_dir, 'genutils.c'),
    os.path.join(libmseed_dir, 'gmtime64.c'),
    os.path.join(libmseed_dir, 'logging.c'),
    os.path.join(libmseed_dir, 'lookup.c'),
    os.path.join(libmseed_dir, 'msio.c'),
    os.path.join(libmseed_dir, 'msrutils.c'),
    os.path.join(libmseed_dir, 'pack.c'),
    os.path.join(libmseed_dir, 'packdata.c'),
    os.path.join(libmseed_dir, 'parseutils.c'),
    os.path.join(libmseed_dir, 'selection.c'),
    os.path.join(libmseed_dir, 'tracelist.c'),
    os.path.join(libmseed_dir, 'unpack.c'),
    os.path.join(libmseed_dir, 'unpackdata.c'),
    os.path.join(libmseed_dir, 'yyjson.c'),
]

# Compiler and linker flags
extra_compile_args = ['-std=c99', '-O2', '-Wall', '-DLIBMSEED_STATIC']
extra_link_args = ['-lm']

# Include directories
include_dirs = [libmseed_dir]

# Define the extension module
ext_modules = [
    Extension(
        name="cymseed3",
        sources=["cymseed3.pyx"] + libmseed_sources,
        include_dirs=include_dirs,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )
]

setup(
    name="cymseed3",
    version="0.1.4",
    description="A Python package for reading and writing miniseed files using libmseed 3.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Valerio Poggi",
    author_email="vpoggi@ogs.it",
    url="https://github.com/shakelab/cymseed3",
    license="GPLv3",
    ext_modules=cythonize(ext_modules),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: C",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Software Development :: Libraries",
        "Topic :: Scientific/Engineering"
    ],
    python_requires='>=3.6',
    setup_requires=[
        'setuptools>=42',
        'cython>=0.29',
    ],
    install_requires=[
        # Add other dependencies your package might have
    ],
    keywords='miniseed libmseed seismic data processing',
    zip_safe=False
)




# ****************************************************************************
#
# Copyright (C) 2019-2024, ShakeLab Developers.
# This file is part of ShakeLab.
#
# ShakeLab is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# ShakeLab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# with this download. If not, see <http://www.gnu.org/licenses/>
#
# ****************************************************************************

# cymseed3.pxd
# cython: language_level=3

cdef extern from "stdint.h":

    ctypedef signed char int8_t
    ctypedef unsigned char uint8_t
    ctypedef signed short int int16_t
    ctypedef unsigned short int uint16_t
    ctypedef signed int int32_t
    ctypedef unsigned int uint32_t
    ctypedef signed long long int64_t
    ctypedef unsigned long long uint64_t
    ctypedef int64_t nstime_t

cdef extern from "libmseed.h":

    cdef int MINRECLEN

    cdef int MS_ENDOFFILE        # End of file reached return value
    cdef int MS_NOERROR          # No error
    cdef int MS_GENERROR         # Generic unspecified error
    cdef int MS_NOTSEED          # Data not SEED
    cdef int MS_WRONGLENGTH      # Length of data read was not correct
    cdef int MS_OUTOFRANGE       # SEED record length out of range
    cdef int MS_UNKNOWNFORMAT    # Unknown data encoding format
    cdef int MS_STBADCOMPFLAG    # Steim, invalid compression flag(s)
    cdef int MS_INVALIDCRC       # Invalid CRC

    cdef int MSF_UNPACKDATA      # Unpack data samples
    cdef int MSF_SKIPNOTDATA     # Skip input that cannot be identified as miniSEED
    cdef int MSF_VALIDATECRC     # Validate CRC (if version 3)
    cdef int MSF_PNAMERANGE      # Parse and utilize byte range from path name suffix
    cdef int MSF_ATENDOFFILE     # Reading routine is at the end of the file
    cdef int MSF_SEQUENCE        # UNSUPPORTED: Maintain a record-level sequence number
    cdef int MSF_FLUSHDATA       # Pack all available data even if final record would not be filled
    cdef int MSF_PACKVER2        # Pack as miniSEED version 2 instead of 3
    cdef int MSF_RECORDLIST      # Build a ::MS3RecordList for each ::MS3TraceSeg
    cdef int MSF_MAINTAINMSTL    # Do not modify a trace list when packing

    cdef int DE_TEXT             # Text encoding (UTF-8)
    cdef int DE_INT16            # 16-bit integer
    cdef int DE_INT32            # 32-bit integer
    cdef int DE_FLOAT32          # 32-bit float (IEEE)
    cdef int DE_FLOAT64          # 64-bit float (IEEE)
    cdef int DE_STEIM1           # Steim-1 compressed integers
    cdef int DE_STEIM2           # Steim-2 compressed integers
    cdef int DE_GEOSCOPE24       # [Legacy] GEOSCOPE 24-bit integer
    cdef int DE_GEOSCOPE163      # [Legacy] GEOSCOPE 16-bit gain ranged, 3-bit exponent
    cdef int DE_GEOSCOPE164      # [Legacy] GEOSCOPE 16-bit gain ranged, 4-bit exponent
    cdef int DE_CDSN             # [Legacy] CDSN 16-bit gain ranged
    cdef int DE_SRO              # [Legacy] SRO 16-bit gain ranged
    cdef int DE_DWWSSN           # [Legacy] DWWSSN 16-bit gain ranged

    ctypedef struct MS3Record:
        const char     *record
        int32_t         reclen
        uint8_t         swapflag
        char            sid[64]
        uint8_t         formatversion
        uint8_t         flags
        nstime_t        starttime
        double          samprate
        int16_t         encoding
        uint8_t         pubversion
        int64_t         samplecnt
        uint32_t        crc
        uint16_t        extralength
        uint32_t        datalength
        char           *extra
        void           *datasamples
        uint64_t        datasize
        int64_t         numsamples
        char            sampletype

    ctypedef struct MS3RecordPtr:
        char           *bufferptr
        #FILE          *fileptr
        char           *filename
        int64_t         fileoffset
        MS3Record      *msr
        nstime_t        endtime
        uint32_t        dataoffset
        void           *prvtptr
        MS3RecordPtr   *next

    ctypedef struct MS3RecordList:
        uint64_t        recordcnt
        MS3RecordPtr   *first
        MS3RecordPtr   *last

    ctypedef struct MS3TraceSeg:
        nstime_t        starttime
        nstime_t        endtime
        double          samprate
        int64_t         samplecnt
        void           *datasamples
        uint64_t        datasize
        int64_t         numsamples
        char            sampletype
        void           *prvtptr
        MS3RecordList  *recordlist
        MS3TraceSeg    *prev
        MS3TraceSeg    *next

    ctypedef struct MS3TraceID:
        char            sid[64]
        uint8_t         pubversion
        nstime_t        earliest
        nstime_t        latest
        void           *prvtptr
        uint32_t        numsegments
        MS3TraceSeg    *first
        MS3TraceSeg    *last
        MS3TraceID     *next[8]
        uint8_t         height

    ctypedef struct MS3TraceList:
        uint32_t        numtraceids
        MS3TraceID      traces
        uint64_t        prngstate

    ctypedef struct MS3Tolerance:
        pass

    MS3Record* msr3_init(
        MS3Record *msr
        )

    void msr3_free(
        MS3Record **ppmsr
        )

    MS3TraceList* mstl3_init (
        MS3TraceList *mstl
        )

    void mstl3_free (
        MS3TraceList **ppmstl,
        int8_t freeprvtptr
        )

    int msr3_parse(
        char* buffer,
        uint64_t recbuflen,
        MS3Record **ppmsr,
        uint32_t dataflag,
        int8_t verbose
        )

    #int msr3_unpack_data(
    #    MS3Record *msr,
    #    int8_t verbose
    #    )

    #int msr3_pack(
    #    const MS3Record *msr,
    #    void (*record_handler) (char *, int, void *),
    #    void *handlerdata,
    #    int64_t *packedsamples,
    #    uint32_t flags,
    #    int8_t verbose
    #    )

    void msr3_print(
        const MS3Record *msr,
        int8_t details
        )

    int64_t mstl3_pack(
        MS3TraceList *mstl,
        void (*record_handler) (char *, int, void *),
        void *handlerdata,
        int reclen,
        int8_t encoding,
        int64_t *packedsamples,
        uint32_t flags,
        int8_t verbose,
        char *extra
        )

    MS3TraceSeg* mstl3_addmsr_recordptr(
        MS3TraceList *mstl,
        const MS3Record *msr,
        MS3RecordPtr **recordptr,
        int8_t splitversion,
        int8_t mergegaps,
        uint32_t flags,
        const MS3Tolerance *tolerance
        )

    int ms_nstime2time(
        int64_t nstime,
        uint16_t *year,
        uint16_t *yday,
        uint8_t *hour,
        uint8_t *min,
        uint8_t *sec,
        uint32_t *nsec
        )

    char* ms_nstime2timestr(
        int64_t nstime,
        char *timestr,
        int timeformat,
        int subsecond
        )

    int64_t ms_time2nstime(
        int year,
        int yday,
        int hour,
        int min,
        int sec,
        uint32_t nsec
        )

    int64_t ms_timestr2nstime(
        const char *timestr
        )

    int ms_sid2nslc(
        const char *sid,
        char *net,
        char *sta,
        char *loc,
        char *chan
        )

    int ms_nslc2sid(
        char *sid,
        int sidlen,
        uint16_t flags,
        const char *net,
        const char *sta,
        const char *loc,
        const char *chan
        )
